#pragma once
#include "IOValues/timer_io_values.hpp"

namespace JabyEngine {
    namespace Timer_IO {
        using namespace Timer_IO_Values;

        using CounterModeIO   = IOPort<Timer_IO_Values::CounterMode>;
        using CounterTargetIO = IOPort<Timer_IO_Values::CounterTarget>;
        using CounterValueIO  = IOPort<Timer_IO_Values::CounterValue>;

        #pragma pack(push, 1)
        struct Counter {
            CounterValueIO  value;
            CounterModeIO   mode;
            CounterTargetIO target;
            uint32_t        unused;

            inline uint16_t get_current_value() const {
                return this->value.read().get(CounterValue::Value);
            }

            inline void set_target_value(uint16_t value) {
                this->target.write(CounterTarget{0}.set_range(CounterTarget::CounterTargetValue, value));
            }

            inline void set_mode(CounterMode mode) {
                this->mode.write(mode);
            }
        };
        #pragma pack(pop)

        #pragma pack(push, 1)
        struct Counter0 : public Counter {
            struct SyncMode {
                static constexpr auto Zero_At_Hblank                      = CounterMode::SyncMode.with(1u);
                static constexpr auto Pause_During_Hblank                 = CounterMode::SyncMode.with(0u);
                static constexpr auto Zero_At_Hblank_Pause_Outside_Hblank = CounterMode::SyncMode.with(2u);
                static constexpr auto Pause_Until_Hblank_Then_Freerun     = CounterMode::SyncMode.with(3u);
            };

            struct Source {
                static constexpr auto System_Clock     = CounterMode::ClockSource.with(0u);
                static constexpr auto Dot_Clock        = CounterMode::ClockSource.with(1u);
                static constexpr auto System_Clock_Too = CounterMode::ClockSource.with(2u);
                static constexpr auto Dot_Clock_Too    = CounterMode::ClockSource.with(3u);
            };
        };

        struct Counter1 : public Counter {
            struct SyncMode {
                static constexpr auto Pause_During_Vblank                 = CounterMode::SyncMode.with(0u);
                static constexpr auto Zero_At_Vblank                      = CounterMode::SyncMode.with(1u);
                static constexpr auto Zero_At_Vblank_Pause_Outside_Vblank = CounterMode::SyncMode.with(2u);
                static constexpr auto Pause_Until_Vblank_Then_FreeRun     = CounterMode::SyncMode.with(3u);
            };

            struct Source {
                static constexpr auto System_Clock     = CounterMode::ClockSource.with(0u);
                static constexpr auto Hblank           = CounterMode::ClockSource.with(1u);
                static constexpr auto System_Clock_Too = CounterMode::ClockSource.with(2u);
                static constexpr auto Hblank_Too       = CounterMode::ClockSource.with(3u);   
            };
        };

        struct Counter2 : public Counter {
            struct SyncMode {
                static constexpr auto Stop_Counter     = CounterMode::SyncMode.with(0u);
                static constexpr auto FreeRun          = CounterMode::SyncMode.with(1u);
                static constexpr auto FreeRun_Too      = CounterMode::SyncMode.with(2u);
                static constexpr auto Stop_Counter_Too = CounterMode::SyncMode.with(3u);
            };

            struct Source {
                static constexpr auto System_Clock           = CounterMode::ClockSource.with(0u);
                static constexpr auto System_Clock_Too       = CounterMode::ClockSource.with(1u);
                static constexpr auto System_Clock_Div_8     = CounterMode::ClockSource.with(2u);
                static constexpr auto System_Clock_Div_8_Too = CounterMode::ClockSource.with(3u);
            };
        };
        #pragma pack(pop)

        static constexpr uintptr_t counter_base_adr(size_t ID) {
            return (0x1F801100 + (ID*0x10));
        }

        // TODO: Improve this to actually use it for measurement - if possible
        static auto& Counter0 = __new_declare_io_value(struct Counter0, counter_base_adr(0));
        static auto& Counter1 = __new_declare_io_value(struct Counter1, counter_base_adr(1));
        static auto& Counter2 = __new_declare_io_value(struct Counter2, counter_base_adr(2));
    }
}